/******************************************************************************
 * resecure - resecure file permissions after system upgrade
 * Copyright (C) 2000  Bill Marquette <billm@danger.ms>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include "resecure.h"
#include "config.h"
#include "trace.h"
#include "rmd160.h"
#include "hashtest.h"
#include "dllist.h"

FILE *dbfd;   /* Database file */
FILE *conffd; /* Config file   */

void resecurefs(void);
void create_db(void);
void print_mode(mode_t st_mode);
char *get_mode(mode_t st_mode);
void print_flags(unsigned int st_flags);
void print_owner(uid_t st_uid, gid_t st_gid);
int set_owner(char *filename, uid_t st_uid, gid_t st_gid);
int set_mode(char *filename, mode_t st_mode);
int set_flags(char *filename, unsigned int st_flags);
void usage(void);
void version(void);

int cflag = 0; /* create_db                 */
int yflag = 0; /* yes to all questions      */
int fflag = 0; /* fix permission difference */
int quiet = 0; /* Make program less chatty  */
  
int main(int argc, char **argv)
{
  extern char *optarg;
  extern int optind;
  extern int USE_TRACE;
  int ch;


  while ((ch = getopt(argc, argv, "CFVqv")) != -1) {
    switch (ch) {
      case 'C': 
	cflag = 1;
	break;
      case 'F':
	fflag = 1;
	break;
      case 'V':
	version();
	exit(1);
      case 'q': 
	quiet = 1;
	break;
      case 'v':
	USE_TRACE++;
	printf("Verbosity at level: %d\n", USE_TRACE);
	break;
      case 'y':
	yflag = 1;
	break;
      default:
	usage();
	exit(1);
    }
  }
  if ((optind == 1) && (argc > 1))
  {
    usage();
    exit(1);
  }
  argc -= optind;
  argv += optind;
  
  if (cflag)
  {
    create_db();
    exit(1);
  } else resecurefs();

  return(1);
}


/******************************************************************************
 * function: resecure_fs()
 * purpose: main resecuring function
 *****************************************************************************/
void resecurefs(void) {
  char *filename;
  file_struct db;    /* file database */
  struct stat sbcmp; /* stat(2) struct for comparison, pulled from db */
  struct stat sb;    /* stat(2) struct */

  unsigned int j;
  byte *hash;
  char *hashstr;
  char dskhash[42];

  TRACE(4,"[function: resecurefs()]\n");

  errno = 0; /* Clear errno before we open database */
  if ((dbfd = fopen(DBFILE, "r")) == NULL)
  {
    fprintf(stderr, "Couldn't open database file for read.\n");
    exit(errno);
  }

  if ((conffd = fopen(CONFFILE, "r")) == NULL)
  {
    fprintf(stderr, "Couldn't open config file for read.\n");
    exit(errno);
  }
  
  while(fread(&db, sizeof(file_struct), 1, dbfd))
  {
    /* Copy stat struct for easier use */
    sbcmp = db.sb;

    filename = (char *)malloc(sizeof(char)*(strlen(db.filename)+strlen(db.dirname)));
    sprintf(filename, "%s",db.filename);
    TRACEV(3, ("dirname: [%s] filename: [%s]\n", db.dirname, db.filename));
    TRACEV(1,("Current file: %s\n", filename));
    if (stat(filename, &sb) == 0)
    {

/* RMD160 hashes */
      hashstr = malloc(sizeof(char)*3);

      /* Get hash from current file on disk */
      hash = RMDbinary((byte *) filename);
      for (j=0; j<RMDsize/8; j++)
      {
	sprintf(hashstr, "%02x", hash[j]);
	dskhash[j*2] = hashstr[0];
	dskhash[(j*2)+1] = hashstr[1];
      }
      dskhash[(j*2)+2] = '\0'; /* NUL terminate string */

      if (strcmp(db.hash, dskhash) != 0) 
      {
	printf("WARNING!!! FILE HASH HAS CHANGED!!!\n");
	printf("Filename: %s\n", filename); 
	printf("Hash on disk: ");
	printf("%s \n", dskhash);
	
	printf("Hash on file: ");
	printf("%s \n", db.hash);
	printf("--------\n");
      }
/*      sprintf(fb.hash, "%s", dskhash); */
/* END RMD160 hash */


      if ((sbcmp.st_flags != sb.st_flags) || (sbcmp.st_mode != sb.st_mode))
      {
	printf("WARNING!!! FILE ATTRIBUTES HAVE BEEN MODIFIED!!!\n");
	printf("Filename: %s\n", filename); 
	printf("Current attributes:");
	print_owner(sb.st_uid, sb.st_gid);
	print_mode(sb.st_mode);
	print_flags(sb.st_flags);
	
	printf("--------\nAttributes on file:\n");
	print_owner(sbcmp.st_uid, sbcmp.st_gid);
	print_mode(sbcmp.st_mode);
	print_flags(sbcmp.st_flags);
	
	if (fflag)
	{
	  printf("Changing file attributes to match what's in database...");
	  if (!set_owner(filename, sbcmp.st_uid, sbcmp.st_gid)) printf(".");
	  else
	    fprintf(stderr, "%s: %s\n", filename, strerror(errno));
	  if (S_ISLNK(sb.st_mode) == 0)
	  {
	    if (!set_mode(filename, sbcmp.st_mode)) printf(".");
	    else
	      fprintf(stderr, "%s: %s\n", filename, strerror(errno));
	  }
	  if (!set_flags(filename, sbcmp.st_flags)) printf(".");
	  else
	    fprintf(stderr, "%s: %s\n", filename, strerror(errno));
	}
	printf("\n------\n\n");
      }
    }  
    free(filename);
  }
  fclose(dbfd);
  fclose(conffd);
}


/******************************************************************************
 * function: create_db()
 * purpose: create database file
 *****************************************************************************/

void create_db(void)
{
  char filename[FILENAME_MAX+1]; /* Hold full filename  */
  char dirname[FILENAME_MAX+1];  /* Hold directory name */

  DIR *dirp;
  struct dirent *dp;

  NodePtr filenode = NULL;
  ListPtr filelist = NULL;
  struct stat sb; /* stat struct from disk        */
  file_struct db; /* database struct w/ filenames */

  unsigned int j;
  byte *hash;
  char *hashstr;

  TRACE(4,"[function create_db()]\n");

  /* Open the database file */
  errno = 0;
  dbfd = fopen(DBFILE, "w");
  if (dbfd == NULL)
  {
    fprintf(stderr, "Error creating database file!\n");
    exit(errno);
  }

  /* Open the config file */
  errno = 0;
  conffd = fopen(CONFFILE, "r");
  if (conffd == NULL)
  {
    fprintf(stderr, "The configuration file doesn't exist!\n");
    exit(errno);
  }

  /* Get directory from conf file */
  while (fgets(dirname, FILENAME_MAX, conffd) != NULL)
  {
    filenode = NULL; /* Set pointers to null */
    filelist = NULL; /* Set pointers to null */

    /* remove the newline */
    dirname[strlen(dirname)-1] = '\0';

    /* Make sure the directory exists and is a directory before going farther */
    if ((stat(dirname,&sb) != 0) && (S_ISDIR(sb.st_mode))) break;
    
    dirp = opendir(dirname);
    if (dirp)
      while ((dp = readdir(dirp)) != NULL)
      {
	filelist = add_to_list((void *)dp->d_name, strlen(dp->d_name)+1, filenode, filelist);
	filenode = filelist->current;
      }


    set_list_to_head(filelist); /* Go back to head of file list */
    
    while ((filelist = get_next_node(filelist)) != NULL)
    {
      filenode = filelist->current;
      snprintf(filename, sizeof(filename)-1, "%s/%s", dirname, (char *)filenode->data);
      
      if (!quiet) printf("File: %s\n", filename);
      
      /* Stat file */
      if (stat(filename, &sb) == 0)
      {
	if(!quiet) {
	  print_owner(sb.st_uid, sb.st_gid);
	  print_mode(sb.st_mode);
	  print_flags(sb.st_flags);
	}

	db.sb = sb;
	strncpy(db.filename, filename, sizeof(db.filename)-1);
	strncpy(db.dirname, dirname, sizeof(db.dirname)-1);
	
/* RMD160 hashes */
	hashstr = malloc(sizeof(char)*3);
	
	hash = RMDbinary((byte *) filename);
	if (!quiet) printf("Hash: ");
	for (j=0; j<RMDsize/8; j++)
	{
	  sprintf(hashstr, "%02x", hash[j]);
	  db.hash[j*2] = hashstr[0];
	  db.hash[(j*2)+1] = hashstr[1];
	}
	db.hash[(j*2)+2] = '\0'; /* NUL terminate string */
	
	if (!quiet) printf("%s \n", db.hash);
/* END RMD160 hash */
	
	fwrite(&db, sizeof(file_struct), 1, dbfd);
      }
    }

    free_list(filelist); /* Free memory used by linked list */

  } /* while(directory in config file) */

  fclose(dbfd);
  fclose(conffd);
}


/******************************************************************************
 * function: print_mode()
 * purpose: print file mode similar to how ls does
 *****************************************************************************/

void print_mode(mode_t st_mode)
{
  TRACE(4,"[function: print_mode()]\n");

  printf("File mode: %s\n", get_mode(st_mode));
}


/******************************************************************************
 * function: get_mode()
 * purpose: return file mode in pretty print format
 *****************************************************************************/

char *get_mode(mode_t st_mode)
{
  static char mode[9];
  char tmode; /* tmp var to hold mode character */

  TRACE(4,"[function: get_mode()]\n");

  /* User */
  if (S_ISXUSR(st_mode)) tmode = 'x';
  else tmode =  '-';
  mode[0] = tmode;

  if (S_ISRUSR(st_mode)) tmode = 'r';
  else tmode = '-';
  mode[1] = tmode;

  if (S_ISWUSR(st_mode)) tmode = 'w';
  else tmode = '-';
  mode[2] = tmode;

  /* Group */
  if (S_ISXGRP(st_mode)) tmode = 'x';
  else tmode = '-';
  mode[3] = tmode;

  if (S_ISRGRP(st_mode)) tmode = 'r';
  else tmode = '-';
  mode[4] = tmode;

  if (S_ISWGRP(st_mode)) tmode = 'w';
  else tmode = '-';
  mode[5] = tmode;

  /* Other */
  if (S_ISXOTH(st_mode)) tmode = 'x';
  else tmode = '-';
  mode[6] = tmode;

  if (S_ISROTH(st_mode)) tmode = 'r';
  else tmode = '-';
  mode[7] = tmode;

  if (S_ISWOTH(st_mode)) 
    tmode = 'w';
  else tmode = '-';
  mode[8] = tmode;

/* Add other modes later +s +g +t */
  if (S_ISSUID(st_mode)) mode[0] = 's';
  if (S_ISSGID(st_mode)) mode[3] = 's';
  if (S_ISSVTX(st_mode)) mode[8] = 't';

  return(mode);
}


/******************************************************************************
 * function: print_flags
 * purpose: print flags on files
 *****************************************************************************/

void print_flags(unsigned int st_flags)
{
  TRACE(4,"[function: print_flags()]\n");

  printf("File flags: ");
  if (S_ISUF_NODUMP(st_flags)) printf("dump ");
  if (S_ISSF_ARCHIVED(st_flags)) printf("archived ");
  if (S_ISUF_IMMUTABLE(st_flags)) printf("user_immutable ");
  if (S_ISSF_IMMUTABLE(st_flags)) printf("root_immutable ");
  if (S_ISUF_APPEND(st_flags)) printf("user_append_only ");
  if (S_ISSF_APPEND(st_flags)) printf("root_append_only ");
  if (S_ISUF_OPAQUE(st_flags)) printf("opaque ");
  printf("\n");
}


/******************************************************************************
 * function: print_owner()
 * purpose: pretty print owner
 *****************************************************************************/

void print_owner(uid_t st_uid, gid_t st_gid)
{
  TRACE(4,"[function: print_owner()]\n");

  printf("UID: [%d]  GID: [%d]\n", st_uid, st_gid);
}


/******************************************************************************
 * function: set_owner()
 * purpose: wrap around chown
 *****************************************************************************/

int set_owner(char *filename, uid_t st_uid, gid_t st_gid)
{
  int retval;

  TRACE(4,"[function: set_owner()]\n");
  TRACEV(1,("File: [%s] UID: [%d]  GID: [%d]\n", filename, st_uid, st_gid));

  errno = 0;
  if ((retval = chown(filename, st_uid, st_gid)) == -1)
    fprintf(stderr, "Error #%d changing owner or group for file: %s\n", errno, filename);

  return(retval);
}


/******************************************************************************
 * function: set_mode()
 * purpose: wrap around chmod
 *****************************************************************************/

int set_mode(char *filename, mode_t st_mode)
{
  int retval;

  TRACE(4,"[function: set_mode()]\n");
  TRACEV(1,("File: [%s] Mode: %d\n", filename, st_mode));

  errno = 0;
  if ((retval = chmod(filename, st_mode)) == -1)
    fprintf(stderr, "Error #%d changing permissions for file: %s\n", errno, filename);

  return(retval);
}


/******************************************************************************
 * function: set_flags()
 * purpose: wrap around chflags
 *****************************************************************************/

int set_flags(char *filename, unsigned int st_flags)
{
  int retval;

  TRACE(4,"[function: set_flags()]\n");
  TRACEV(1,("File: [%s] Mode: %d\n", filename, st_flags));

  if ((retval = chflags(filename, st_flags)) == -1)
    fprintf(stderr, "Error #%d changing flags for file: %s\n", errno, filename);

  return(retval);
}


/******************************************************************************
 * function: usage()
 * purpose: give brief overview of program usage
 *****************************************************************************/

void usage(void)
{
  TRACE(4,"[function: usage()]\n");

  printf("usage: resecure [-C] [-V]\n");
  printf("\t-C\tCreate database\n");
  printf("\t-F\tFix file permissions to match database\n");
  printf("\t-V\tVersion\n");
  printf("\t-q\tQuiet\n");
  printf("\t-v\tVerbose - each additional use adds verbosity\n");
}


/******************************************************************************
 * function: version()
 * purpose: display program version
 *****************************************************************************/

void version(void)
{
  TRACE(4,"[function: version()]\n");

  printf("resecure -- version %s %s  %s\n", VERSION, COPYRIGHT, AUTHOR);
}
